from __future__ import annotations
import csv
from fractions import Fraction
import pandas as pd
from pydantic import BaseModel, validator
from typing import Optional
import numpy as np
from .ballot import Ballot


class PreferenceProfile(BaseModel):
    """
    PreferenceProfile class, contains ballots and candidates for a given election.

    Args:
        ballots (list[Ballot]): List of ``Ballot`` objects.
        candidates (list, optional): List of candidate strings. Defaults to None.


    Attributes:
        ballots (list[Ballot]): List of ``Ballot`` objects.
        candidates (list, optional): List of candidate strings.
        df (pandas.DataFrame): pandas ``DataFrame`` used to store ballots.
    """

    ballots: list[Ballot] = []
    candidates: Optional[list] = None
    df: pd.DataFrame = pd.DataFrame()

    @validator("candidates", allow_reuse=True)
    def cands_must_be_unique(cls, candidates: list) -> list:
        if not len(set(candidates)) == len(candidates):
            raise ValueError("All candidates must be unique")
        return candidates

    def get_ballots(self) -> list[Ballot]:
        """
        Returns:
            list: List of ``Ballot`` objects.
        """
        return self.ballots[:]

    def get_candidates(self, received_votes: Optional[bool] = True) -> list:
        """
        Args:
            received_votes (bool, optional): If True, only return candidates that received votes.
                Defaults to True.
        Returns:
            list: List of candidates.
        """

        if received_votes or not self.candidates:
            unique_cands: set = set()
            for ballot in self.ballots:
                unique_cands.update(*ballot.ranking)

            return list(unique_cands)
        else:
            return self.candidates

    # can also cache
    def num_ballots(self) -> Fraction:
        """
        Counts number of ballots based on assigned weight.

        Returns:
            Fracton: Number of ballots cast.
        """
        num_ballots = Fraction(0)
        for ballot in self.ballots:
            num_ballots += ballot.weight

        return num_ballots

    def to_dict(self, standardize: bool = False) -> dict:
        """
        Converts profile to dictionary with keys = rankings and
        values = corresponding total weights.

        Args:
            standardize (bool, optional): If True, divides the weight of each ballot by the total
                weight. Defaults to False.

        Returns:
            dict: A dictionary with ranking (keys) and corresponding total weights (values).
        """
        num_ballots = self.num_ballots()
        di: dict = {}
        for ballot in self.ballots:
            rank_tuple = tuple(next(iter(item)) for item in ballot.ranking)
            if standardize:
                weight = ballot.weight / num_ballots
            else:
                weight = ballot.weight
            if rank_tuple not in di.keys():
                di[rank_tuple] = weight
            else:
                di[rank_tuple] += weight
        return di

    class Config:
        arbitrary_types_allowed = True

    def to_csv(self, fpath: str):
        """
        Saves PreferenceProfile to CSV.

        Args:
            fpath (str): Path to the saved csv.
        """
        with open(fpath, "w", newline="") as csvfile:
            fieldnames = ["weight", "ranking"]
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for ballot in self.ballots:
                writer.writerow({"weight": ballot.weight, "ranking": ballot.ranking})

    def _create_df(self) -> pd.DataFrame:
        """
        Creates pandas DataFrame for display and building plots.

        Returns:
            pandas.DataFrame: Data frame of ballots with weight and percent columns.
        """
        weights = []
        ballots = []
        for ballot in self.ballots:
            part = []
            for ranking in ballot.ranking:
                if len(ranking) == 1:
                    part.append(list(ranking)[0])

                else:
                    part.append(f"{ranking} (Tie)")

            ballots.append(tuple(part))
            weights.append(ballot.weight)

        df = pd.DataFrame({"Ballots": ballots, "Weight": weights})

        try:
            df["Percent"] = df["Weight"] / df["Weight"].sum()
        except ZeroDivisionError:
            df["Percent"] = np.nan

        # fill nans with zero for edge cases
        df["Percent"] = df["Percent"].fillna(0.0)

        def format_as_percent(frac):
            return f"{float(frac):.2%}"

        df["Percent"] = df["Percent"].apply(format_as_percent)
        return df.reset_index(drop=True)

    def head(
        self,
        n: int,
        sort_by_weight: Optional[bool] = True,
        percents: Optional[bool] = False,
        totals: Optional[bool] = False,
    ) -> pd.DataFrame:
        """
        Displays top-n ballots in profile.

        Args:
            n (int): Number of ballots to view.
            sort_by_weight (bool, optional): If True, rank ballot from most to least votes.
                Defaults to True.
            percents (bool, optional): If True, show voter share for a given ballot.
                Defaults to False.
            totals (bool, optional): If True, show total values for Percent and Weight.
                Defaults to False.

        Returns:
            pd.DataFrame: A dataframe with top-n ballots.
        """
        if self.df.empty:
            self.df = self._create_df()

        if sort_by_weight:
            df = (
                self.df.sort_values(by="Weight", ascending=False)
                .head(n)
                .reset_index(drop=True)
            )
        else:
            df = self.df.head(n).reset_index(drop=True)

        if totals:
            df = self._sum_row(df)

        if not percents:
            return df.drop(columns="Percent")

        return df

    def tail(
        self,
        n: int,
        sort_by_weight: Optional[bool] = True,
        percents: Optional[bool] = False,
        totals: Optional[bool] = False,
    ) -> pd.DataFrame:
        """
        Displays bottom-n ballots in profile.

        Args:
            n (int): Number of ballots to view.
            sort_by_weight (bool, optional): If True, rank ballot from least to most votes.
                Defaults to True.
            percents (bool, optional): If True, show voter share for a given ballot.
                Defaults to False.
            totals (bool, optional): If True, show total values for Percent and Weight.
                Defaults to False.

        Returns:
            A data frame with bottom-n ballots.
        """

        if self.df.empty:
            self.df = self._create_df()

        if sort_by_weight:
            df = self.df.sort_values(by="Weight", ascending=True)
            df["New Index"] = [x for x in range(len(self.df) - 1, -1, -1)]
            df = df.set_index("New Index").head(n)
            df.index.name = None

        else:
            df = self.df.iloc[::-1].head(n)

        if totals:
            df = self._sum_row(df)

        if not percents:
            return df.drop(columns="Percent")

        return df

    def __str__(self) -> str:
        # Displays top 15 cast ballots or entire profile

        if self.df.empty:
            self.df = self._create_df()

        if len(self.df) < 15:
            return self.head(n=len(self.df), sort_by_weight=True).to_string(
                index=False, justify="justify"
            )

        print(
            f"PreferenceProfile too long, only showing 15 out of {len(self.df) } rows."
        )
        return self.head(n=15, sort_by_weight=True).to_string(
            index=False, justify="justify"
        )

    # set repr to print outputs
    __repr__ = __str__

    def condense_ballots(self) -> PreferenceProfile:
        """
        Groups ballots by rankings and updates weights.

        Returns:
            PreferenceProfile: A PreferenceProfile object with condensed ballot list.
        """
        ranking_to_index: dict = {}
        weight_accumulator = {}

        for ballot in self.ballots:
            if ballot.ranking not in ranking_to_index:
                ranking_to_index[ballot.ranking] = len(ranking_to_index)
                weight_accumulator[ballot.ranking] = Fraction(0)

            weight_accumulator[ballot.ranking] += ballot.weight

        new_ballot_list = [
            Ballot(ranking=tuple(map(frozenset, ranking)), weight=weight)
            for ranking, weight in weight_accumulator.items()
        ]

        condensed_profile = PreferenceProfile(ballots=new_ballot_list)
        return condensed_profile

    def __eq__(self, other):
        if not isinstance(other, PreferenceProfile):
            return False
        pp_1 = self.condense_ballots()
        pp_2 = other.condense_ballots()
        for b in pp_1.ballots:
            if b not in pp_2.ballots:
                return False
        for b in pp_2.ballots:
            if b not in pp_1.ballots:
                return False
        return True

    def _sum_row(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Computes sum total for weight and percent column
        """

        def format_as_float(percent_str):
            return float(percent_str.split("%")[0])

        sum_row = {
            "Ballot": "",
            "Weight": f'{df["Weight"].sum()} out of {self.num_ballots()}',
            "Percent": f'{df["Percent"].apply(format_as_float).sum():.2f} out of 100%',
        }

        df.loc["Totals"] = sum_row  # type: ignore

        return df.fillna("")

    def __add__(self, other):
        """
        Add two PreferenceProfiles by combining their ballot lists.
        """
        if isinstance(other, PreferenceProfile):
            ballots = self.ballots + other.ballots
            pp = PreferenceProfile(ballots=ballots)
            pp.condense_ballots()
            return pp
        else:
            raise TypeError(
                "Unsupported operand type. Must be an instance of PreferenceProfile."
            )



from copy import deepcopy
from fractions import Fraction
from functools import reduce
from itertools import groupby
from typing import Callable

from .pref_profile import PreferenceProfile
from .ballot import Ballot


def remove_empty_ballots(
    pp: PreferenceProfile, keep_candidates: bool = False
) -> PreferenceProfile:
    """
    Removes empty ballots from a PreferenceProfile.

    Args:
        pp (PreferenceProfile): A PreferenceProfile to clean.
        keep_candidates (bool, optional): If True, keep all of the candidates
            from the original PreferenceProfile in the returned PreferenceProfile, even if
            they got no votes. Defaults to False.

    Returns:
        PreferenceProfile: A cleaned PreferenceProfile.
    """

    ballots_nonempty = [
        deepcopy(ballot) for ballot in pp.get_ballots() if ballot.ranking
    ]
    if keep_candidates:
        old_cands = deepcopy(pp.get_candidates())
        pp_clean = PreferenceProfile(ballots=ballots_nonempty, candidates=old_cands)
    else:
        pp_clean = PreferenceProfile(ballots=ballots_nonempty)
    return pp_clean


def clean_profile(
    pp: PreferenceProfile, clean_ballot_func: Callable[[Ballot], Ballot]
) -> PreferenceProfile:
    """
    Allows user-defined cleaning rules for PreferenceProfile. Input function
    that applies modification to a single ballot.

    Args:
        pp (PreferenceProfile): A PreferenceProfile to clean.
        clean_ballot_func (Callable[[Ballot], Ballot]): Function that
            takes a ``Ballot`` and returns a cleaned ``Ballot``.

    Returns:
        PreferenceProfile: A cleaned ``PreferenceProfile``.
    """

    # apply cleaning function to clean all ballots
    if clean_ballot_func is not None:
        cleaned = map(clean_ballot_func, pp.ballots)
    # group ballots that have the same ranking after cleaning
    grouped_ballots = [
        list(result)
        for key, result in groupby(cleaned, key=lambda ballot: ballot.ranking)
    ]
    # merge ballots in the same groups
    new_ballots = [merge_ballots(b) for b in grouped_ballots]
    return PreferenceProfile(ballots=new_ballots)


def merge_ballots(ballots: list[Ballot]) -> Ballot:
    """
    Takes a list of ballots with the same ranking and merges them into one ballot.

    Args:
        ballots (list[Ballot]): A list of ballots to deduplicate.

    Returns:
        Ballot: A ballot with the same ranking and aggregated weight and voters.
    """
    weight = sum(b.weight for b in ballots)
    ranking = ballots[0].ranking
    voters_to_merge = [b.voter_set for b in ballots if b.voter_set]
    voter_set = None
    if len(voters_to_merge) > 0:
        voter_set = reduce(lambda b1, b2: b1.union(b2), voters_to_merge)
        voter_set = set(voter_set)
    return Ballot(ranking=ranking, voter_set=voter_set, weight=Fraction(weight))


def deduplicate_profiles(pp: PreferenceProfile) -> PreferenceProfile:
    """
    Given a PreferenceProfile, cleans its ballots by removing a candidate from the remainder
    of the ballot if they have already appeared.

    Args:
        pp (PreferenceProfile): A PreferenceProfile to clean.

    Returns:
        PreferenceProfile: A cleaned PreferenceProfile where each ballot has a ranking
        with no repeated candidates.
    """

    def deduplicate_ballots(ballot: Ballot) -> Ballot:
        """
        Takes a ballot and deduplicates its rankings.

        Args:
            ballot (Ballot): a ballot with duplicates in its ranking.

        Returns:
            Ballot: a ballot without duplicates.
        """
        ranking = ballot.ranking
        dedup_ranking = []
        for cand in ranking:
            if cand in ranking and cand not in dedup_ranking:
                dedup_ranking.append(cand)
        new_ballot = Ballot(
            id=ballot.id,
            weight=Fraction(ballot.weight),
            ranking=tuple(dedup_ranking),
            voter_set=ballot.voter_set,
        )
        return new_ballot

    pp_clean = clean_profile(pp=pp, clean_ballot_func=deduplicate_ballots)
    return pp_clean


def remove_noncands(
    profile: PreferenceProfile, non_cands: list[str]
) -> PreferenceProfile:
    """
    Removes user-assigned non-candidates from ballots, deletes ballots
    that are empty as a result of the removal.

    Args:
        profile (PreferenceProfile): A PreferenceProfile to clean.
        non_cands (list[str]): A list of non-candidates to be removed.

    Returns:
        PreferenceProfile: A profile with non-candidates removed.
    """

    def remove_from_ballots(ballot: Ballot, non_cands: list[str]) -> Ballot:
        """
        Removes non-candidiates from ballot objects.

        Args:
            ballot (Ballot): a ballot to be cleaned.
            non_cands (list[str]): a list of candidates to remove.

        Returns:
            Ballot: returns a cleaned Ballot.
        """
        # TODO: adjust so string and list of strings are acceptable inputes

        to_remove = []
        for item in non_cands:
            to_remove.append({item})

        ranking = ballot.ranking
        clean_ranking = []
        for cand in ranking:
            if cand not in to_remove and cand not in clean_ranking:
                clean_ranking.append(cand)

        clean_ballot = Ballot(
            id=ballot.id,
            ranking=tuple(clean_ranking),
            weight=Fraction(ballot.weight),
            voter_set=ballot.voter_set,
        )

        return clean_ballot

    cleaned = [
        remove_from_ballots(ballot, non_cands)
        for ballot in profile.ballots
        if remove_from_ballots(ballot, non_cands).ranking
    ]
    grouped_ballots = [
        list(result)
        for key, result in groupby(cleaned, key=lambda ballot: ballot.ranking)
    ]
    # merge ballots in the same groups
    new_ballots = [merge_ballots(b) for b in grouped_ballots]
    return PreferenceProfile(ballots=new_ballots)



from matplotlib import pyplot as plt  # type: ignore
from votekit.pref_profile import PreferenceProfile
from votekit.utils import first_place_votes, mentions, COLOR_LIST, borda_scores


def plot_summary_stats(
    profile: PreferenceProfile, stat: str, multi_color: bool = True, title: str = ""
):
    """
    Plots histogram of election summary statistics.

    Args:
        profile (PreferenceProfile): A ``PreferenceProfile`` to visualize.
        stat (str): 'first place votes', 'mentions', or 'borda'.
        multi_color (bool, optional): If the bars should be multicolored. Defaults to True.
        title (str, optional): Title for the figure. Defaults to empty string.

    Returns:
        Figure: A ``matplotlib`` figure with the visualization.
    """
    stats = {
        "first place votes": first_place_votes,
        "mentions": mentions,
        "borda": borda_scores,
    }

    stat_func = stats[stat]
    data: dict = stat_func(profile)  # type: ignore

    if multi_color:
        colors = COLOR_LIST[: len(list(data.keys()))]
    else:
        colors = [COLOR_LIST[-1]]

    fig, ax = plt.subplots()

    candidates = profile.get_candidates(received_votes=False)
    y_data = [data[c] for c in candidates]

    ax.bar(candidates, y_data, color=colors, width=0.35)
    ax.set_xlabel("Candidates")
    ax.set_ylabel("Frequency")

    if title:
        ax.set_title(title)

    return fig



from fractions import Fraction
import itertools as it
import numpy as np
from typing import Callable, Optional, Union
from functools import lru_cache

from ..models import Election
from ..election_state import ElectionState
from ..graphs.pairwise_comparison_graph import PairwiseComparisonGraph
from ..pref_profile import PreferenceProfile
from .transfers import fractional_transfer
from ..utils import (
    compute_votes,
    remove_cand,
    scores_into_set_list,
    tie_broken_ranking,
    elect_cands_from_set_ranking,
    first_place_votes,
    compute_scores_from_vector,
    validate_score_vector,
    borda_scores,
    ballots_by_first_cand,
)


class STV(Election):
    """
    Class for multi-winner STV elections.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        transfer (Callable): Transfer method (e.g. fractional transfer).
        seats (int): Number of seats to be elected.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        transfer (Callable): Transfer method (e.g. fractional transfer).
        seats (int): Number of seats to be elected.
        quota (str): Formula to calculate quota.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
        threshold (int): Threshold number of votes to be elected.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        transfer: Callable,
        seats: int,
        quota: str = "droop",
        ballot_ties: bool = True,
        tiebreak: Union[str, Callable] = "random",
    ):
        # let parent class handle the og profile and election state
        super().__init__(profile, ballot_ties)

        self.transfer = transfer
        self.seats = seats
        self.tiebreak = tiebreak
        self.quota = quota.lower()
        self.threshold = self.get_threshold()

    # can cache since it will not change throughout rounds
    def get_threshold(self) -> int:
        """
        Calculates threshold required for election.

        Returns:
            int: Value of the threshold.
        """
        quota = self.quota
        if quota == "droop":
            return int(self._profile.num_ballots() / (self.seats + 1) + 1)
        elif quota == "hare":
            return int(self._profile.num_ballots() / self.seats)
        else:
            raise ValueError("Misspelled or unknown quota type")

    def next_round(self) -> bool:
        """
        Determines if the number of seats has been met to call an election.

        Returns:
            bool: True if number of seats has not been met, False otherwise.
        """
        cands_elected = 0
        for s in self.state.winners():
            cands_elected += len(s)
        return cands_elected < self.seats

    def run_step(self) -> ElectionState:
        """
        Simulates one round an STV election.

        Returns:
           ElectionState: An ElectionState object for a round.
        """
        remaining = self.state.profile.get_candidates()
        ballots = self.state.profile.get_ballots()
        round_votes, plurality_score = compute_votes(remaining, ballots)

        elected = []
        eliminated = []

        # if number of remaining candidates equals number of remaining seats,
        # everyone is elected
        if len(remaining) == self.seats - len(
            [c for s in self.state.winners() for c in s]
        ):
            elected = [{cand} for cand, _ in round_votes]
            remaining = []
            ballots = []

        # elect all candidates who crossed threshold
        elif round_votes[0].votes >= self.threshold:
            # partition ballots by first place candidate
            cand_to_ballot = ballots_by_first_cand(remaining, ballots)
            new_ballots = []
            for candidate, votes in round_votes:
                if votes >= self.threshold:
                    elected.append({candidate})
                    remaining.remove(candidate)
                    # only transfer on ballots where winner is first
                    new_ballots += self.transfer(
                        candidate,
                        cand_to_ballot[candidate],
                        plurality_score,
                        self.threshold,
                    )

            # add in remaining ballots where non-winners are first
            for cand in remaining:
                new_ballots += cand_to_ballot[cand]

            # remove winners from all ballots
            ballots = remove_cand([c for s in elected for c in s], new_ballots)

        # since no one has crossed threshold, eliminate one of the people
        # with least first place votes
        else:
            lp_candidates = [
                candidate
                for candidate, votes in round_votes
                if votes == round_votes[-1].votes
            ]

            if isinstance(self.tiebreak, str):
                lp_cand = tie_broken_ranking(
                    ranking=[set(lp_candidates)],
                    profile=self.state.profile,
                    tiebreak=self.tiebreak,
                )[-1]
            else:
                lp_cand = self.tiebreak(
                    ranking=[set(lp_candidates)], profile=self.state.profile
                )[-1]

            eliminated.append(lp_cand)
            ballots = remove_cand(lp_cand, ballots)
            remaining.remove(next(iter(lp_cand)))

        # sorts remaining based on their current first place votes
        _, score_dict = compute_votes(remaining, ballots)
        remaining = scores_into_set_list(score_dict, remaining)

        # sort candidates by vote share if multiple are elected
        if len(elected) >= 1:
            elected = scores_into_set_list(
                plurality_score, [c for s in elected for c in s]
            )

        # Make sure list-of-sets have non-empty elements
        elected = [s for s in elected if s != set()]
        eliminated = [s for s in eliminated if s != set()]

        self.state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated_cands=eliminated,
            remaining=remaining,
            scores=score_dict,
            profile=PreferenceProfile(ballots=ballots),
            previous=self.state,
        )
        return self.state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Runs complete STV election.

        Returns:
            ElectionState: An ElectionState object with results for a complete election.
        """
        if not self.next_round():
            raise ValueError(
                f"Length of elected set equal to number of seats ({self.seats})"
            )

        while self.next_round():
            self.run_step()

        return self.state


class Limited(Election):
    """
    Elects m candidates with the highest k-approval scores.
    The k-approval score of a candidate is equal to the number of voters who
    rank this candidate among their k top ranked candidates.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        k (int): The number of top ranked candidates to consider as approved.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        k (int): The number of top ranked candidates to consider as approved.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        k: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.k = k
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts Limited election in which m candidates are elected based
        on approval scores.

        Returns:
           ElectionState: An ElectionState object for a Limited election.
        """
        profile = self.state.profile
        candidates = profile.get_candidates()
        candidate_approvals = {c: Fraction(0) for c in candidates}

        for ballot in profile.get_ballots():
            # First we have to determine which candidates are approved
            # i.e. in first k ranks on a ballot
            approvals = []
            for i, cand_set in enumerate(ballot.ranking):
                # If list of total candidates before and including current set
                # are less than seat count, all candidates are approved
                if len(list(it.chain(*ballot.ranking[: i + 1]))) < self.k:
                    approvals.extend(list(cand_set))
                # If list of total candidates before current set
                # are greater than seat count, no candidates are approved
                elif len(list(it.chain(*ballot.ranking[:i]))) > self.k:
                    approvals.extend([])
                # Else we know the cutoff is in the set, we compute and randomly
                # select the number of candidates we can select
                else:
                    accepted = len(list(it.chain(*ballot.ranking[:i])))
                    num_to_allow = self.k - accepted
                    approvals.extend(
                        np.random.choice(list(cand_set), num_to_allow, replace=False)
                    )

            # Add approval votes equal to ballot weight (i.e. number of voters with this ballot)
            for cand in approvals:
                candidate_approvals[cand] += ballot.weight

        # Order candidates by number of approval votes received
        ranking = scores_into_set_list(candidate_approvals)

        if isinstance(self.tiebreak, str):
            ranking = tie_broken_ranking(
                ranking=ranking, profile=self.state.profile, tiebreak=self.tiebreak
            )
        else:
            ranking = self.tiebreak(ranking=ranking, profile=self.state.profile)

        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )
        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated_cands=eliminated,
            remaining=list(),
            scores=candidate_approvals,
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return self.state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete Limited election.

        Returns:
            ElectionState: An ElectionState object with results for a complete election.
        """
        self.run_step()
        return self.state


class Bloc(Limited):
    """
    Elects m candidates with the highest m-approval scores. Specific case of Limited election
    where k = m.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, seats, seats, ballot_ties, tiebreak)


class SNTV(Limited):
    """
    Single nontransferable vote (SNTV): Elects k candidates with the highest
    Plurality scores. Equivalent to Limited with k=1.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, seats, 1, ballot_ties, tiebreak)


class Plurality(SNTV):
    """
    Simulates a single or multi-winner plurality election. Wrapper for SNTV.
    """


class SNTV_STV_Hybrid(Election):
    """
    Election method that first runs SNTV to a cutoff number of candidates, then runs STV to
    pick a committee with a given number of seats.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        transfer (Callable): Transfer method (e.g. fractional transfer).
        r1_cutoff (int): First round cutoff value.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        transfer (Callable): Transfer method (e.g. fractional transfer).
        r1_cutoff (int): First round cutoff value.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.


    """

    def __init__(
        self,
        profile: PreferenceProfile,
        transfer: Callable,
        r1_cutoff: int,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.transfer = transfer
        self.r1_cutoff = r1_cutoff
        self.seats = seats
        self.tiebreak = tiebreak
        self.stage = "SNTV"  # SNTV, switches to STV, then Complete

    def run_step(self, stage: str) -> ElectionState:
        """
        Simulates one round an SNTV_STV election.

        Args:
            stage (str): Stage of the hybrid election, can be "SNTV" or "STV".

        Returns:
            ElectionState: An ElectionState object for a given round.
        """
        profile = self.state.profile

        new_state = None
        if stage == "SNTV":
            round_state = SNTV(
                profile=profile, seats=self.r1_cutoff, tiebreak=self.tiebreak
            ).run_election()

            # The STV election will be run on the new election state
            # Therefore we should not add any winners, but rather
            # set the SNTV winners as remaining candidates and update pref profiles
            new_profile = PreferenceProfile(
                ballots=remove_cand(
                    set().union(*round_state.eliminated_cands), profile.get_ballots()
                )
            )
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=list(),
                eliminated_cands=round_state.eliminated_cands,
                remaining=[set(new_profile.get_candidates())],
                profile=new_profile,
                scores=round_state.get_scores(round_state.curr_round),
                previous=self.state,
            )
        elif stage == "STV":
            round_state = STV(
                profile=profile,
                transfer=self.transfer,
                seats=self.seats,
                tiebreak=self.tiebreak,
            ).run_election()

            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=round_state.winners(),
                eliminated_cands=round_state.eliminated(),
                remaining=round_state.remaining,
                scores=round_state.get_scores(round_state.curr_round),
                profile=round_state.profile,
                previous=self.state,
            )

        # Update election stage to cue next run step
        if stage == "SNTV":
            self.stage = "STV"
        elif stage == "STV":
            self.stage = "Complete"

        self.state = new_state  # type: ignore
        return new_state  # type: ignore

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Runs complete SNTV_STV election.

        Returns:
            ElectionState: An ElectionState object with results for a complete election.
        """
        while self.stage != "Complete":
            self.run_step(self.stage)
        return self.state  # type: ignore


class TopTwo(Election):
    """
    Eliminates all but the top two plurality vote getters, and then
    conducts a runoff between them, reallocating other ballots.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        ballot_ties: bool = True,
        tiebreak: Union[str, Callable] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a TopTwo election for one seat with a cutoff of 2 for the runoff.

        Returns:
            ElectionState: An ElectionState object for the TopTwo election.
        """
        hybrid_equivalent = SNTV_STV_Hybrid(
            profile=self.state.profile,
            transfer=fractional_transfer,
            r1_cutoff=2,
            seats=1,
            tiebreak=self.tiebreak,
        )
        outcome = hybrid_equivalent.run_election()
        self.state = outcome
        return outcome

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete TopTwo election.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        self.run_step()
        return self.state


class DominatingSets(Election):
    """
    Finds tiers of candidates by dominating set, which is a set of candidates
    such that every candidate in the set wins head to head comparisons against
    candidates outside of it. Elects all candidates in the top tier.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
    """

    def __init__(self, profile: PreferenceProfile, ballot_ties: bool = True):
        super().__init__(profile, ballot_ties)

    def run_step(self) -> ElectionState:
        """
        Conducts a complete DominatingSets election as it is not a round-by-round
        system.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        pwc_graph = PairwiseComparisonGraph(self.state.profile)
        dominating_tiers = pwc_graph.dominating_tiers()
        if len(dominating_tiers) == 1:
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=list(),
                eliminated_cands=dominating_tiers,
                remaining=list(),
                scores=pwc_graph.pairwise_dict,
                profile=PreferenceProfile(),
                previous=self.state,
            )
        else:
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=[set(dominating_tiers[0])],
                eliminated_cands=dominating_tiers[1:],
                remaining=list(),
                scores=pwc_graph.pairwise_dict,
                profile=PreferenceProfile(),
                previous=self.state,
            )
        self.state = new_state
        return new_state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete DominatingSets election.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        self.run_step()
        return self.state


class CondoBorda(Election):
    """
    Finds tiers of candidates by dominating set, which is a set of candidates
    such that every candidate in the set wins head to head comparisons against
    candidates outside of it. Elects as many candidates as specified, in order from
    first to last dominating set. If the number of seats left is smaller than the next
    dominating set, CondoBorda breaks ties between candidates with Borda.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a complete Conda-Borda election as it is not a round-by-round
        system.

        Returns:
            ElectionState: An `ElectionState` object for a complete election.
        """
        pwc_graph = PairwiseComparisonGraph(self.state.profile)
        dominating_tiers = pwc_graph.dominating_tiers()

        if isinstance(self.tiebreak, str):
            ranking = tie_broken_ranking(
                ranking=dominating_tiers, profile=self.state.profile, tiebreak="borda"
            )
        else:
            ranking = self.tiebreak(
                ranking=dominating_tiers, profile=self.state.profile
            )

        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )

        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated_cands=eliminated,
            remaining=list(),
            scores=pwc_graph.pairwise_dict,
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return new_state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete Conda-Borda election.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        self.run_step()
        return self.state


class SequentialRCV(Election):
    """
    Class to conduct Sequential RCV election, in which votes are not transferred
    after a candidate has reached threshold, or been elected.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self, old_profile: PreferenceProfile) -> ElectionState:
        """
        Simulates a single step of the sequential RCV contest or a full
        IRV election run on the current set of candidates.

        Returns:
            ElectionState: An ElectionState object.
        """
        old_election_state = self.state

        IRVrun = STV(
            old_profile,
            transfer=(lambda winner, ballots, votes, threshold: ballots),
            seats=1,
            tiebreak=self.tiebreak,
        )
        old_election = IRVrun.run_election()
        elected_cand = old_election.winners()[0]

        # Removes elected candidate from Ballot List
        updated_ballots = remove_cand(elected_cand, old_profile.get_ballots())

        # Updates profile with removed candidates
        updated_profile = PreferenceProfile(ballots=updated_ballots)

        self.state = ElectionState(
            curr_round=old_election_state.curr_round + 1,
            elected=[elected_cand],
            profile=updated_profile,
            previous=old_election_state,
            scores=first_place_votes(updated_profile),
            remaining=old_election.remaining,
        )
        return self.state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete sequential RCV contest.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        old_profile = self._profile
        elected = []  # type: ignore
        seqRCV_step = self.state

        while len(elected) < self.seats:
            seqRCV_step = self.run_step(old_profile)
            elected.append(seqRCV_step.elected)
            old_profile = seqRCV_step.profile
        return seqRCV_step


class Borda(Election):
    """
    Positional voting system that assigns a decreasing number of points to
    candidates based on order and a score vector. The conventional score
    vector is :math:`(n, n-1, \dots, 1)`, where `n` is the number of candidates.
    If a ballot is incomplete, the remaining points of the score vector
    are evenly distributed to the unlisted candidates (see ``borda_scores`` function in ``utils``).

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        score_vector (list[Fraction], optional): Weights assigned to candidate ranking.
            Defaults to :math:`(n,n-1,\dots,1)`.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        score_vector (list[Fraction], optional): Weights assigned to candidate ranking.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        score_vector: Optional[list[Fraction]] = None,
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak
        self.score_vector = score_vector

    def run_step(self) -> ElectionState:
        """
        Simulates a complete Borda contest as Borda is not a round-by-round
        system.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        borda_dict = borda_scores(
            profile=self.state.profile, score_vector=self.score_vector
        )

        ranking = scores_into_set_list(borda_dict)

        if isinstance(self.tiebreak, str):
            ranking = tie_broken_ranking(
                ranking=ranking, profile=self.state.profile, tiebreak=self.tiebreak
            )
        else:
            ranking = self.tiebreak(ranking=ranking, profile=self.state.profile)

        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )

        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated_cands=eliminated,
            remaining=list(),
            scores=borda_dict,
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return new_state

    @lru_cache
    def run_election(self) -> ElectionState:
        """
        Simulates a complete Borda contest.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        self.run_step()
        return self.state


class IRV(STV):
    """
    A class for conducting IRV elections, which are mathematically equivalent to STV for one seat.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        quota (str): Formula to calculate quota.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
        threshold (int): Threshold number of votes to be elected.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        quota: str = "droop",
        ballot_ties: bool = True,
        tiebreak: Union[Callable, str] = "random",
    ):
        # let parent class handle the construction
        super().__init__(
            profile=profile,
            ballot_ties=ballot_ties,
            seats=1,
            tiebreak=tiebreak,
            quota=quota,
            transfer=fractional_transfer,
        )


class HighestScore(Election):
    """
    Conducts an election based on points from score vector.
    Chooses the m candidates with highest scores.
    Ties are broken by randomly permuting the tied candidates.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        score_vector (list[float]): List of floats where `i`th entry denotes the number of points
            given to candidates ranked in position `i`.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        score_vector (list[float]): List of floats where `i`th entry denotes the number of points
            given to candidates ranked in position `i`.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        score_vector: list[float],
        tiebreak: Union[Callable, str] = "random",
        ballot_ties: bool = False,
    ):
        super().__init__(profile, ballot_ties)
        # check for valid score vector
        validate_score_vector(score_vector)

        self.seats = seats
        self.score_vector = score_vector
        self.tiebreak = tiebreak

    def run_step(self):
        """
        Simulates a complete Borda contest as Borda is not a round-by-round
        system.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        # a dictionary whose keys are candidates and values are scores
        vote_tallies = compute_scores_from_vector(
            profile=self.state.profile, score_vector=self.score_vector
        )

        # translate scores into ranking of candidates, tie break
        ranking = scores_into_set_list(score_dict=vote_tallies)

        if isinstance(self.tiebreak, str):
            untied_ranking = tie_broken_ranking(
                ranking=ranking, profile=self.state.profile, tiebreak=self.tiebreak
            )
        else:
            untied_ranking = self.tiebreak(ranking=ranking, profile=self.state.profile)

        elected, eliminated = elect_cands_from_set_ranking(
            ranking=untied_ranking, seats=self.seats
        )

        self.state = ElectionState(
            curr_round=1,
            elected=elected,
            eliminated_cands=eliminated,
            remaining=[],
            profile=self.state.profile,
            previous=self.state,
        )
        return self.state

    @lru_cache
    def run_election(self):
        """
        Simulates a complete Borda contest.

        Returns:
            ElectionState: An ElectionState object for a complete election.
        """
        self.run_step()
        return self.state


class Cumulative(HighestScore):
    """
    Voting system where voters are allowed to vote for candidates with multiplicity.
    Each ranking position should have one candidate, and every candidate ranked will receive
    one point, i.e., the score vector is :math:`(1,\dots,1)`.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        seats (int): Number of seats to be elected.
        ballot_ties (bool, optional): Resolves input ballot ties if True, else assumes ballots have
            no ties. Defaults to True.
        tiebreak (Union[str, Callable], optional): Resolves procedural and final ties by specified
            tiebreak. Can either be a custom tiebreak function or a string. Supported strings are
            given in ``tie_broken_ranking`` documentation. The custom function must take as
            input two named parameters; ``ranking``, a list-of-sets ranking of candidates and
            ``profile``, the original ``PreferenceProfile``. It must return a list-of-sets
            ranking of candidates with no ties. Defaults to random tiebreak.

    Attributes:
        _profile (PreferenceProfile):   PreferenceProfile to run election on.
        state (ElectionState): Current state of the election.
        seats (int): Number of seats to be elected.
        tiebreak (Union[str, Callable]): Resolves procedural and final ties by specified
            tiebreak.
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: Union[str, Callable] = "random",
    ):
        longest_ballot = 0
        for ballot in profile.ballots:
            if len(ballot.ranking) > longest_ballot:
                longest_ballot = len(ballot.ranking)

        score_vector = [1.0 for _ in range(longest_ballot)]
        super().__init__(
            profile=profile,
            ballot_ties=ballot_ties,
            score_vector=score_vector,
            seats=seats,
            tiebreak=tiebreak,
        )
